/*
 * This file is part of the OdinMS Maple Story Server
 * Copyright (C) 2008 Patrick Huy <patrick.huy@frz.cc>
 * Matthias Butz <matze@odinms.de>
 * Jan Christian Meyer <vimes@odinms.de>
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation version 3 as published by
 * the Free Software Foundation. You may not use, modify or distribute
 * this program under any other version of the GNU Affero General Public
 * License.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package tools;

import org.apache.commons.lang3.StringUtils;

public class StringUtil{

	public static String getLeftPaddedInt(int in, int length){
		return String.format("%0" + length + "d", in);
	}

	/**
	 * Gets a string padded from the left to <code>length</code> by
	 * <code>padchar</code>.
	 *
	 * @param in The input string to be padded.
	 * @param padchar The character to pad with.
	 * @param length The length to pad to.
	 * @return The padded string.
	 */
	public static String getLeftPaddedStr(String in, char padchar, int length){
		StringBuilder builder = new StringBuilder(length);
		for(int x = in.length(); x < length; x++){
			builder.append(padchar);
		}
		builder.append(in);
		return builder.toString();
	}

	/**
	 * Gets a string padded from the right to <code>length</code> by
	 * <code>padchar</code>.
	 *
	 * @param in The input string to be padded.
	 * @param padchar The character to pad with.
	 * @param length The length to pad to.
	 * @return The padded string.
	 */
	public static String getRightPaddedStr(String in, char padchar, int length){
		StringBuilder builder = new StringBuilder(in);
		for(int x = in.length(); x < length; x++){
			builder.append(padchar);
		}
		return builder.toString();
	}

	/**
	 * Joins an array of strings starting from string <code>start</code> with a
	 * space.
	 *
	 * @param arr The array of strings to join.
	 * @param start Starting from which string.
	 * @return The joined strings.
	 */
	public static String joinStringFrom(String arr[], int start){
		return joinStringFrom(arr, start, " ");
	}

	/**
	 * Joins an array of strings starting from string <code>start</code> with
	 * <code>sep</code> as a seperator.
	 *
	 * @param arr The array of strings to join.
	 * @param start Starting from which string.
	 * @return The joined strings.
	 */
	public static String joinStringFrom(String arr[], int start, String sep){
		StringBuilder builder = new StringBuilder();
		for(int i = start; i < arr.length; i++){
			builder.append(arr[i]);
			if(i != arr.length - 1){
				builder.append(sep);
			}
		}
		return builder.toString();
	}

	/**
	 * Makes an enum name human readable (fixes spaces, capitalization, etc)
	 *
	 * @param enumName The name of the enum to neaten up.
	 * @return The human-readable enum name.
	 */
	public static String makeEnumHumanReadable(String enumName){
		StringBuilder builder = new StringBuilder(enumName.length() + 1);
		String[] words = enumName.split("_");
		for(String word : words){
			if(word.length() <= 2){
				builder.append(word); // assume that it's an abbrevation
			}else{
				builder.append(word.charAt(0));
				builder.append(word.substring(1).toLowerCase());
			}
			builder.append(' ');
		}
		return builder.substring(0, enumName.length());
	}

	/**
	 * Counts the number of <code>chr</code>'s in <code>str</code>.
	 *
	 * @param str The string to check for instances of <code>chr</code>.
	 * @param chr The character to check for.
	 * @return The number of times <code>chr</code> occurs in <code>str</code>.
	 */
	public static int countCharacters(String str, char chr){
		int ret = 0;
		for(int i = 0; i < str.length(); i++){
			if(str.charAt(i) == chr){
				ret++;
			}
		}
		return ret;
	}

	/**
	 * Gets the current time in a readable format
	 * 
	 * @param startMillis
	 * @param endMillis
	 * @return A user-friendly String
	 */
	public static String getReadableMillis(long startMillis, long endMillis){
		StringBuilder sb = new StringBuilder();
		double elapsedSeconds = (endMillis - startMillis) / 1000.0;
		int elapsedSecs = ((int) elapsedSeconds) % 60;
		int elapsedMinutes = (int) (elapsedSeconds / 60.0);
		int elapsedMins = elapsedMinutes % 60;
		int elapsedHrs = elapsedMinutes / 60;
		int elapsedHours = elapsedHrs % 24;
		int elapsedDays = elapsedHrs / 24;
		if(elapsedDays > 0){
			boolean mins = elapsedHours > 0;
			sb.append(elapsedDays);
			sb.append(" day").append(elapsedDays > 1 ? "s" : "").append(mins ? ", " : ".");
			if(mins){
				boolean secs = elapsedMins > 0;
				if(!secs){
					sb.append("and ");
				}
				sb.append(elapsedHours);
				sb.append(" hour").append(elapsedHours > 1 ? "s" : "").append(secs ? ", " : ".");
				if(secs){
					boolean millis = elapsedSecs > 0;
					if(!millis){
						sb.append("and ");
					}
					sb.append(elapsedMins);
					sb.append(" minute").append(elapsedMins > 1 ? "s" : "").append(millis ? ", " : ".");
					if(millis){
						sb.append("and ");
						sb.append(elapsedSecs);
						sb.append(" second").append(elapsedSecs > 1 ? "s" : "").append(".");
					}
				}
			}
		}else if(elapsedHours > 0){
			boolean mins = elapsedMins > 0;
			sb.append(elapsedHours);
			sb.append(" hour").append(elapsedHours > 1 ? "s" : "").append(mins ? ", " : ".");
			if(mins){
				boolean secs = elapsedSecs > 0;
				if(!secs){
					sb.append("and ");
				}
				sb.append(elapsedMins);
				sb.append(" minute").append(elapsedMins > 1 ? "s" : "").append(secs ? ", " : ".");
				if(secs){
					sb.append("and ");
					sb.append(elapsedSecs);
					sb.append(" second").append(elapsedSecs > 1 ? "s" : "").append(".");
				}
			}
		}else if(elapsedMinutes > 0){
			boolean secs = elapsedSecs > 0;
			sb.append(elapsedMinutes);
			sb.append(" minute").append(elapsedMinutes > 1 ? "s" : "").append(secs ? " " : ".");
			if(secs){
				sb.append("and ");
				sb.append(elapsedSecs);
				sb.append(" second").append(elapsedSecs > 1 ? "s" : "").append(".");
			}
		}else if(elapsedSeconds > 0){
			sb.append((int) elapsedSeconds);
			sb.append(" second").append(elapsedSeconds > 1 ? "s" : "").append(".");
		}else{
			sb.append("None.");
		}
		return sb.toString();
	}

	/**
	 * Calculates the similarity (a number within 0 and 1) between two strings.
	 */
	public static double similarity(String s1, String s2){
		String longer = s1.toLowerCase(), shorter = s2.toLowerCase();
		if(s1.length() < s2.length()){ // longer should always have greater length
			longer = s2;
			shorter = s1;
		}
		int longerLength = longer.length();
		if(longerLength == 0){ return 1.0; /* both strings are zero length */ }
		return (longerLength - StringUtils.getLevenshteinDistance(longer, shorter)) / (double) longerLength;
	}
}
